local addonName, AT = ...
local DEBUG = false
local function _assert(aCnd, aMsg) if DEBUG then assert(aCnd, aMsg) end return aCnd end

-- file import
local Constants = AT.Constants
local FrameMessageBox = AT.FrameMessageBox
local Settings = AT.Settings
local Strings = AT.Strings
local Utils = AT.Utils

-- enum import
local PROFILE_TYPE = AT.Enums.PROFILE_TYPE

--[[
  NOTES:
  - Acts as static "class"
  - is very quick and dirty
]]

-- needs to be a frame for callbacks
local FrameSettings = CreateFrame('frame')
FrameSettings.frame = nil
FrameSettings.categoryName = AT.Constants.ADDON_DISPLAY_NAME
FrameSettings.createdFrames = {}
FrameSettings.framePairs = {}
FrameSettings.frameProfilesHelp = nil

AT.FrameSettings = FrameSettings

local function InvokeCallback(aCallback, aFrame)
  if aCallback ~= nil then
    aCallback(aFrame)
  end
end

local dropdownMenuFramePoint = {
  { text = "Center", value = 'CENTER' },
  { text = "Top Left", value = 'TOPLEFT' },
  { text = "Top", value = 'TOP' },
  { text = "Top Right", value = 'TOPRIGHT' },
  { text = "Right", value = 'RIGHT' },
  { text = "Bottom Right", value = 'BOTTOMRIGHT' },
  { text = "Bottom", value = 'BOTTOM' },
  { text = "Bottom Left", value = 'BOTTOMLEFT' },
  { text = "Left", value = 'LEFT' }
}

local dropdownMenuFrameStrata = {
  { text = "Background", value = 'BACKGROUND' },
  { text = "Low", value = 'LOW' },
  { text = "Medium", value = 'MEDIUM' },
  { text = "High", value = 'HIGH' },
  { text = "Dialog", value = 'DIALOG' },
  { text = "Fullscreen", value = 'FULLSCREEN' },
  { text = "Fullscreen Dialog", value = 'FULLSCREEN_DIALOG' },
  { text = "Tooltip", value = 'TOOLTIP' }
}

local dropdownMenuSoundChannel = {
  { text = "Master", value = 'Master' },
  { text = "SFX", value = 'SFX' },
  { text = "Ambience", value = 'Ambience' },
  { text = "Music", value = 'Music' },
  { text = "Dialog", value = 'Dialog' }
}

local dropdownMenuChatMessage = {
  { text = "None", value = AT.Enums.BUFF_STOLEN_CHAT_MESSAGE.NONE },
  { text = "Only me", value = AT.Enums.BUFF_STOLEN_CHAT_MESSAGE.PRIVATE },
  { text = "Public", value = AT.Enums.BUFF_STOLEN_CHAT_MESSAGE.PUBLIC }
}

local dropdownMenuNpOrientation = {
  { text = "Left", value = AT.Enums.NP_ORIENTATION.LEFT },
  { text = "Center", value = AT.Enums.NP_ORIENTATION.CENTER },
  { text = "Right", value = AT.Enums.NP_ORIENTATION.RIGHT }
}

local dropdownMenuIsBlacklist = {
  { text = "Blacklist", value = 1 },
  { text = "Whitelist", value = 0 }
}

local DEFAULT_WIDGET_SETTINGS = {
  parent = nil,
  point = 'TOPLEFT',
  posX = 20,
  posY = -20,
  scale = 1
}

function FrameSettings:Init(aIsMage)
  if aIsMage then
    self:InitMage()
  else
    self:InitNonMage()
  end
end

function FrameSettings:InitMage()
  -- add listener for profile change
  Settings.OnProfileChanged:Add(self, self.RefreshAllValues, PROFILE_TYPE.GENERAL + PROFILE_TYPE.FILTER)

  -- get resolution
  local resWidth, resHeight = Utils:GetResolution()

  -- reused vars for caching values
  local posX, posX1, posX2, posX3, posY, posY1, posY2, posY3, f, f1, f2, f3

  local widgetSettings

  -- multiple times used vars and functions
  local sliderMaxCenteredFrameX = math.ceil(resWidth / 2)
  local sliderMinCenteredFrameX = sliderMaxCenteredFrameX * -1
  local sliderMaxCenteredFrameY = math.ceil(resHeight / 2)
  local sliderMinCenteredFrameY = sliderMaxCenteredFrameY * -1

  local function Rescan() AT.Main:Rescan() end
  local function UpdateFrameTarget() AT.Main.frameTarget:UpdateSettings() end
  local function UpdateFrameFocus() AT.Main.frameFocus:UpdateSettings() end
  local function UpdateFrameOverview() AT.Main.frameOverview:UpdateSettings() end
  local function UpdateFrameMouseover() AT.Main.frameMouseover:UpdateSettings() end
  local function UpdateAllFrames() UpdateFrameTarget(); UpdateFrameFocus(); UpdateFrameOverview(); UpdateFrameMouseover() end
  local function UpdateNameplates() AT.NameplateFrames:UpdateSettings(); Rescan() end
  local function UpdateCustomFilters() AT.Main:UpdateCustomFilters(); Rescan() end
  local function UpdateTextColors() UpdateAllFrames(); Rescan() end

  -------------
  -- Shared
  -------------

  -- Show test frames
  local function AddTestUtils(aWidgetSettings)
    local paddingX = 10
    local paddingY = -10
    local parent = aWidgetSettings.parent
    local widgetSettings = Utils:TableShallowCopy(DEFAULT_WIDGET_SETTINGS)
    widgetSettings.posX = paddingX
    widgetSettings.posY = paddingY
    local width = 0
    local f = CreateFrame('Frame', nil, aWidgetSettings.parent)
    f:SetBackdrop({
      --bgFile = 'Interface/Tooltips/UI-Tooltip-Background',
      edgeFile = 'Interface/Tooltips/UI-Tooltip-Border',
      tile = true, tileSize = 16, edgeSize = 16,
      insets = { left = 4, right = 4, top = 4, bottom = 4 }
    })
    --f:SetBackdropColor(0, 0.5, 1, 1)
    widgetSettings.parent = f

    -- add headline
    posX = widgetSettings.posX
    self:AddHeadline("Testing", widgetSettings)
    width = math.max(width, self.createdFrames[#self.createdFrames]:GetWidth())
    widgetSettings.posX = posX
    widgetSettings.posY = widgetSettings.posY + 5

    -- add testframes checkbutton
    self:AddCheckButton('testFramesEnabled', "Show test frames", nil, widgetSettings):HookScript('OnClick', function(aCb) 
      AT.Main:SetTestFramesEnabled(aCb:GetChecked())
      self:RefreshAllValues()
    end)
    width = math.max(width, self:LastCreatedFrame():GetWidth() + getglobal(self:LastCreatedFrame():GetName().. "Text"):GetWidth())
    -- add mouseover tooltip checkbutton
    -- TODO: if scaling happens more often handle the position adjustment in the component
    widgetSettings.posX = widgetSettings.posX + 15
    widgetSettings.posY = widgetSettings.posY - 10
    widgetSettings.scale = DEFAULT_WIDGET_SETTINGS.scale * 0.8
    self:AddCheckButton('testFramesShowMouseover', "Mouseover tooltip", "Show test frame for mouseover tooltip", widgetSettings):HookScript('OnClick', function()
      Rescan()
      self:RefreshAllValues()
    end)
    width = math.max(width, self:LastCreatedFrame():GetWidth() + getglobal(self:LastCreatedFrame():GetName().. "Text"):GetWidth())
    -- add nameplate icons checkbutton
    self:AddCheckButton('testFramesShowNameplateIcons', "Nameplate icons", "Show test icons on nameplates", widgetSettings):HookScript('OnClick', function()
      Rescan()
      AT.Main:OnNameplateTestIconsEnabledChanged()
    end)
    width = math.max(width, self:LastCreatedFrame():GetWidth() + getglobal(self:LastCreatedFrame():GetName().. "Text"):GetWidth())
    widgetSettings.posX = widgetSettings.posX - 15
    widgetSettings.posY = widgetSettings.posY + 10
    widgetSettings.scale = DEFAULT_WIDGET_SETTINGS.scale

    -- add test spots button
    widgetSettings.posY = widgetSettings.posY + 8
    self:AddButton("Test Spots", width - 10, 200, 22, widgetSettings, function() FrameMessageBox:Show("Test Spots", Strings.testSpots) end)
    width = math.max(width, self:LastCreatedFrame():GetWidth() + paddingX)
    widgetSettings.posY = widgetSettings.posY - 18

    -- increase frame size
    local height = math.abs(widgetSettings.posY) - math.abs(paddingY)
    f:SetSize(width + paddingX, height)
    f:ClearAllPoints()
    f:SetPoint('BOTTOMRIGHT', parent, 'BOTTOMRIGHT', -3, 3)
    f:Show()
  end

  ---------------------------
  --        Main
  ---------------------------

  -- create frame
  self:EnsureMainFrame()
  -- setup refresh function
  self.frame.refresh = function()
    -- xpcall because okay/cancdel/default/refresh hides errors
    xpcall(function()
      -- refresh all values
      self:RefreshAllValues()

      -- register to update func in order to handle categories that are only headlines
      if not self:GetScript('OnUpdate') then
        self:SetScript('OnUpdate', self.OnUpdate)
      end
    end, geterrorhandler())
  end

  -- setup okay function
  self.frame.okay = function()
    -- xpcall because okay/cancdel/default/refresh hides errors
    xpcall(function()
      -- workaround for the issue that pressing the okay button overrides cvars that 
      -- have been modified for the nameplate test icons since opening the option menu
      if Settings.testFramesEnabled and Settings.testFramesShowNameplateIcons then
        AT.Main:OnNameplateTestIconsEnabledChanged()
      end
    end, geterrorhandler())
  end

  --refresh widget settings
  widgetSettings = Utils:TableShallowCopy(DEFAULT_WIDGET_SETTINGS)
  widgetSettings.parent = self.frame

  ---------------------------
  --        General
  ---------------------------

  _, widgetSettings = self:AddSubCategory("General", true)

  -- general settings
  self:AddHeadline("General", widgetSettings)
  self:AddCheckButton(nil, "Automatically load addon for currently played class", nil, widgetSettings, function(aFrame)
    if aFrame:GetChecked() then
      AT.Main:AddCurrentClassToWhitelist()
    else
      AT.Main:RemoveCurrentClassFromWhitelist()
    end
  end)
  self:LastCreatedFrame():SetChecked(AT.Main:IsCurrentClassWhitelisted())
  self:AddCheckButton('scanOutOfCombat', "Enable scanning out of combat", nil, widgetSettings, Rescan)
  self:AddCheckButton('frameDragWithModifierKey', "Drag frames with hold \"Alt\" key", nil, widgetSettings)
  self:AddDropdown('buffStolenChatMessage', "Spell stolen chat message", nil, dropdownMenuChatMessage, 100, false, widgetSettings)

  -- toogle features
  widgetSettings.posY = widgetSettings.posY - 20
  self:AddHeadline("Features", widgetSettings)
  posY1 = widgetSettings.posY
  self:AddCheckButton('frameTargetEnabled', "Frame for target unit", nil, widgetSettings, Rescan)
  self:AddCheckButton('frameFocusEnabled', "Frame for focus unit", nil, widgetSettings, Rescan)
  self:AddCheckButton('frameMouseoverEnabled', "Mouseover tooltip", nil, widgetSettings, Rescan)
  posY2 = widgetSettings.posY
  widgetSettings.posX = widgetSettings.posX + 200
  widgetSettings.posY = posY1
  self:AddCheckButton('frameOverviewEnabled', "Overview frame", nil, widgetSettings, Rescan)
  self:AddCheckButton('npEnabled', "Nameplate icons", nil, widgetSettings, Rescan)
  widgetSettings.posX = DEFAULT_WIDGET_SETTINGS.posX
  widgetSettings.posY = posY2

  -- other settings hint
  widgetSettings.posY = widgetSettings.posY - 20
  self:AddHeadline("More Settings", widgetSettings)
  self:AddText(Strings.otherSettingsHint, 12, 400, widgetSettings)

  -- bug report hint
  f = self:AddText(Strings.bugReportHint, 11, 400, widgetSettings)
  f:ClearAllPoints()
  f:SetPoint('BOTTOMLEFT', 10, 10)

  -- test utils
  AddTestUtils(widgetSettings, true, true)


  ---------------------------
  --        Filter
  ---------------------------

  _, widgetSettings = self:AddSubCategory("Filter")

  ---------------------------
  --   Filter Built-in
  ---------------------------

  _, widgetSettings = self:AddSubCategory("  Built-in", true)

  self:AddHeadline("Units - All", widgetSettings)
  self:AddCheckButton('filterIgnoreUnitsNotAffectingCombat', "Ignore units not affecting combat", nil, widgetSettings, Rescan)
  widgetSettings.posY = widgetSettings.posY - 20

  self:AddHeadline("Buffs - All", widgetSettings)
  self:AddCheckButton('filterIgnoreGroupBuffs', "Ignore class raid buffs", nil, widgetSettings, Rescan)
  widgetSettings.posY = widgetSettings.posY - 20

  self:AddHeadline("Buffs - Target & Focus", widgetSettings)
  self:AddSlider('filterImportantMinExpirationTime', "Min. expiration time", nil, 0, 30, 0, widgetSettings, UpdateNameplates)
  widgetSettings.posY = widgetSettings.posY - 20

  self:AddHeadline("Buffs - Overview Frame and Nameplates", widgetSettings)
  self:AddSlider('filterOverviewMinExpirationTime', "Min. expiration time", nil, 0, 30, 0, widgetSettings, UpdateNameplates)
  widgetSettings.posY = widgetSettings.posY - 20

  -- frame overview content
  self:AddHeadline("What to display in overview frame", widgetSettings)
  local y = widgetSettings.posY
  local minY = y
  self:AddCheckButton('filterOverviewShowTarget', "Target", nil, widgetSettings, Rescan)
  self:AddCheckButton('filterOverviewShowFocus', "Focus", nil, widgetSettings, Rescan)
  self:AddCheckButton('filterOverviewShowNameplates', "Nameplates", nil, widgetSettings, Rescan)
  minY = math.min(minY, widgetSettings.posY)
  widgetSettings.posX = widgetSettings.posX + 175
  widgetSettings.posY = y
  self:AddCheckButton('filterOverviewShowBosses', "Bosses", nil, widgetSettings, Rescan)
  self:AddCheckButton('filterOverviewShowPetTarget', "Pet target", nil, widgetSettings, Rescan)
  minY = math.min(minY, widgetSettings.posY)
  widgetSettings.posX = widgetSettings.posX + 175
  widgetSettings.posY = y
  self:AddCheckButton('filterOverviewShowGroupMemberTargets', "Group member targets", nil, widgetSettings, Rescan)
  self:AddCheckButton('filterOverviewShowGroupMemberPetTargets', "Group member pet targets", nil, widgetSettings, Rescan)

  ---------------------------
  -- Filter Contextual (CSV)
  ---------------------------

  _, widgetSettings = self:AddSubCategory("  Soft (CSV)")

  self:AddHeadline("Filter: Soft (Comma Seperated Values)", widgetSettings)
  self:AddText(Strings.filterContextualCsv, 12, 582, widgetSettings)

  self:AddDropdown('filterContextualCsvIsBlacklist', "Filter type", nil, dropdownMenuIsBlacklist , 100, true, widgetSettings, UpdateCustomFilters, true, -7)
  posY = widgetSettings.posY
  f = self:AddMultilineEditBox('filterContextualCsv', 582, 300, widgetSettings, UpdateCustomFilters)
  f:ClearAllPoints()
  f:SetPoint('TOPLEFT', widgetSettings.parent, 'TOPLEFT', 20, posY)
  f:SetPoint('BOTTOMRIGHT', widgetSettings.parent, 'BOTTOMRIGHT', -20, 20)

  ---------------------------
  -- Filter Contextual (Lua)
  ---------------------------

  _, widgetSettings = self:AddSubCategory("  Soft (Lua)")

  self:AddHeadline("Filter: Soft (Lua Code)", widgetSettings)
  self:AddText(Strings.filterContextualLua1, 12, 582, widgetSettings)
  self:AddButton("How-to write Lua filters", 0, 200, 22, widgetSettings, function() FrameMessageBox:Show("How-to: Lua Filter", Strings.howToFilterContextualLua) end)
  widgetSettings.posY = widgetSettings.posY - 20
  self:AddText(Strings.filterContextualLua2, 12, 582, widgetSettings)

  posY = widgetSettings.posY
  f = self:AddMultilineEditBox('filterContextualLua', 582, 300, widgetSettings, UpdateCustomFilters)
  f:ClearAllPoints()
  f:SetPoint('TOPLEFT', widgetSettings.parent, 'TOPLEFT', 20, posY)
  f:SetPoint('BOTTOMRIGHT', widgetSettings.parent, 'BOTTOMRIGHT', -20, 20)

  ---------------------------
  --  Filter Global (CSV)
  ---------------------------

  _, widgetSettings = self:AddSubCategory("  Hard (CSV)")

  self:AddHeadline("Filter: Hard (Comma Seperated Values)", widgetSettings)
  self:AddText(Strings.filterGlobalCsv, 12, 582, widgetSettings)

  self:AddDropdown('filterGlobalCsvIsBlacklist', "Filter type", nil, dropdownMenuIsBlacklist , 100, true, widgetSettings, UpdateCustomFilters, true, -7)
  posY = widgetSettings.posY
  f = self:AddMultilineEditBox('filterGlobalCsv', 582, 300, widgetSettings, UpdateCustomFilters)
  f:ClearAllPoints()
  f:SetPoint('TOPLEFT', widgetSettings.parent, 'TOPLEFT', 20, posY)
  f:SetPoint('BOTTOMRIGHT', widgetSettings.parent, 'BOTTOMRIGHT', -20, 20)

  ---------------------------
  --      Appearance
  ---------------------------

  _, widgetSettings = self:AddSubCategory("Appearance")

  ---------------------------
  --       Colors
  ---------------------------

  _, widgetSettings = self:AddSubCategory("  Colors", true)

  -- colors
  self:AddHeadline("Colors", widgetSettings)
  self:AddColorPicker('colorBg', "Background color", nil, true, widgetSettings, UpdateAllFrames)
  self:AddColorPicker('colorBorder', "Border color", nil, true, widgetSettings, UpdateAllFrames)
  self:AddColorPicker('colorHeadline', "Headline color", nil, false, widgetSettings, UpdateTextColors)
  self:AddColorPicker('colorUnitName', "Unit name color", nil, false, widgetSettings, UpdateTextColors)
  self:AddColorPicker('colorUnitOrigin', "Unit origin color", nil, false, widgetSettings, UpdateTextColors)
  self:AddColorPicker('colorBuffName', "Buff name color", nil, false, widgetSettings, UpdateTextColors)
  self:AddColorPicker('colorBuffCount', "Buff stack count color", nil, false, widgetSettings, UpdateTextColors)
  self:AddColorPicker('colorBuffExpTime', "Buff expiration time color", false, nil, widgetSettings, UpdateTextColors)

  -- situation-dependent transparency alpha values
  widgetSettings.posX = widgetSettings.posX + 275
  widgetSettings.posY = DEFAULT_WIDGET_SETTINGS.posY
  self:AddHeadline("Transparency", widgetSettings)
  self:AddSlider('alphaNoMana', "No mana", nil, 0, 1, 2, widgetSettings, UpdateAllFrames)
  self:AddSlider('alphaOutOfRange', "Out of range", nil, 0, 1, 2, widgetSettings, UpdateAllFrames)

  -- test utils
  AddTestUtils(widgetSettings, true)

  ---------------------------
  -- Appearance Nameplates
  ---------------------------

  _, widgetSettings = self:AddSubCategory("  Nameplates")

  -- pos
  self:AddHeadline("Icon Positioning", widgetSettings)
  self:AddDropdown('npIconOrientation', "Orientation", nil, dropdownMenuNpOrientation, 100, false, widgetSettings, UpdateNameplates)
  self:AddSlider('npOffsetX', "Offset X", nil, -100, 100, 0, widgetSettings, UpdateNameplates)
  self:AddSlider('npOffsetY', "Offset Y", nil, -100, 100, 0, widgetSettings, UpdateNameplates)
  self:AddDropdown('npStrata', "Layer", nil, dropdownMenuFrameStrata, 100, false, widgetSettings, UpdateNameplates)

  -- size
  widgetSettings.posX = widgetSettings.posX + 190
  widgetSettings.posY = DEFAULT_WIDGET_SETTINGS.posY
  self:AddHeadline("Icon Size", widgetSettings)
  self:AddSlider('npIconWidth', "Width", "Sets the icon width. Does not include borders.", 4, 30, 0, widgetSettings, UpdateNameplates)
  self:AddSlider('npIconHeight', "Height", "Sets the icon height. Does not include borders.", 4, 30, 0, widgetSettings, UpdateNameplates)
  self:AddSlider('npIconSpacing', "Spacing", "Sets the spacing between two icons", 0, 50, 0, widgetSettings, UpdateNameplates)

  -- borders
  widgetSettings.posX = widgetSettings.posX + 190
  widgetSettings.posY = DEFAULT_WIDGET_SETTINGS.posY
  self:AddHeadline("Border", widgetSettings)
  self:AddSlider('npIconBorderSize', "Icon border size", "Sets the icon border size. This value gets added to the icon size.", 0, 8, 0, widgetSettings, UpdateNameplates)
  self:AddColorPicker('npIconBorderColor', "Icon border color", nil, true, widgetSettings, UpdateNameplates)

  -- test utils
  AddTestUtils(widgetSettings, false, true)

  ---------------------------
  --   Appearance Target
  ---------------------------

  _, widgetSettings = self:AddSubCategory("  Frame Target")

  -- pos
  self:AddHeadline("Positioning", widgetSettings)
  self:AddDropdown('frameTargetAnchor', "Anchor", Strings.tooltipAnchor, dropdownMenuFramePoint, 100, false, widgetSettings, UpdateFrameTarget)
  self:AddDropdown('frameTargetPivot', "Zero point", Strings.tooltipPivot, dropdownMenuFramePoint, 100, false, widgetSettings, UpdateFrameTarget)
  self:AddSlider('frameTargetOffsetX', "Offset X", nil, sliderMinCenteredFrameX, sliderMaxCenteredFrameX, 0, widgetSettings, UpdateFrameTarget)
  self:AddSlider('frameTargetOffsetY', "Offset Y", nil, sliderMinCenteredFrameY, sliderMaxCenteredFrameY, 0, widgetSettings, UpdateFrameTarget)
  self:AddDropdown('frameTargetStrata', "Layer", nil, dropdownMenuFrameStrata, 100, false, widgetSettings, UpdateFrameTarget)

  -- size
  widgetSettings.posX = widgetSettings.posX + 200
  widgetSettings.posY = DEFAULT_WIDGET_SETTINGS.posY
  self:AddHeadline("Size", widgetSettings)
  self:AddSlider('frameTargetScale', "Scale", nil, 0.5, 2, 2, widgetSettings, UpdateFrameTarget)

  -- appearance
  widgetSettings.posX = widgetSettings.posX + 200
  widgetSettings.posY = DEFAULT_WIDGET_SETTINGS.posY
  self:AddHeadline("Appearance", widgetSettings)
  self:AddCheckButton('frameTargetShowBorder', "Show border", nil, widgetSettings, UpdateFrameTarget)

  -- test utils
  AddTestUtils(widgetSettings, true, false)

  ---------------------------
  --   Appearance Focus
  ---------------------------

  _, widgetSettings = self:AddSubCategory("  Frame Focus")

  -- pos
  self:AddHeadline("Positioning", widgetSettings)
  self:AddDropdown('frameFocusAnchor', "Anchor", Strings.tooltipAnchor, dropdownMenuFramePoint, 100, false, widgetSettings, UpdateFrameFocus)
  self:AddDropdown('frameFocusPivot', "Zero point", Strings.tooltipPivot, dropdownMenuFramePoint, 100, false, widgetSettings, UpdateFrameFocus)
  self:AddSlider('frameFocusOffsetX', "Offset X", nil, sliderMinCenteredFrameX, sliderMaxCenteredFrameX, 0, widgetSettings, UpdateFrameFocus)
  self:AddSlider('frameFocusOffsetY', "Offset Y", nil, sliderMinCenteredFrameY, sliderMaxCenteredFrameY, 0, widgetSettings, UpdateFrameFocus)
  self:AddDropdown('frameFocusStrata', "Layer", nil, dropdownMenuFrameStrata, 100, false, widgetSettings, UpdateFrameFocus)

  -- size
  widgetSettings.posX = widgetSettings.posX + 200
  widgetSettings.posY = DEFAULT_WIDGET_SETTINGS.posY
  self:AddHeadline("Size", widgetSettings)
  self:AddSlider('frameFocusScale', "Scale", nil, 0.5, 2, 2, widgetSettings, UpdateFrameFocus)

  -- appearance
  widgetSettings.posX = widgetSettings.posX + 200
  widgetSettings.posY = DEFAULT_WIDGET_SETTINGS.posY
  self:AddHeadline("Appearance", widgetSettings)
  self:AddCheckButton('frameFocusShowBorder', "Show border", nil, widgetSettings, UpdateFrameFocus)

  -- test utils
  AddTestUtils(widgetSettings, true, false)

  ---------------------------
  --  Appearance Overview
  ---------------------------

  _, widgetSettings = self:AddSubCategory("  Frame Overview")

  -- pos
  self:AddHeadline("Positioning", widgetSettings)
  self:AddDropdown('frameOverviewAnchor', "Anchor", Strings.tooltipAnchor, dropdownMenuFramePoint, 100, false, widgetSettings, UpdateFrameOverview)
  self:AddDropdown('frameOverviewPivot', "Zero point", Strings.tooltipPivot, dropdownMenuFramePoint, 100, false, widgetSettings, UpdateFrameOverview)
  self:AddSlider('frameOverviewOffsetX', "Offset X", nil, sliderMinCenteredFrameX, sliderMaxCenteredFrameX, 0, widgetSettings, UpdateFrameOverview)
  self:AddSlider('frameOverviewOffsetY', "Offset Y", nil, sliderMinCenteredFrameY, sliderMaxCenteredFrameY, 0, widgetSettings, UpdateFrameOverview)
  self:AddDropdown('frameOverviewStrata', "Layer", nil, dropdownMenuFrameStrata, 100, false, widgetSettings, UpdateFrameOverview)

  -- size
  widgetSettings.posX = widgetSettings.posX + 200
  widgetSettings.posY = DEFAULT_WIDGET_SETTINGS.posY
  self:AddHeadline("Size", widgetSettings)
  self:AddSlider('frameOverviewScale', "Scale", nil, 0.5, 2, 2, widgetSettings, UpdateFrameOverview)
  self:AddSlider('frameOverviewMinWidth', "Min width", nil, 0, Utils:Round(resWidth / 2), 0, widgetSettings, Rescan)
  self:AddSlider('frameOverviewMaxWidth', "Max width", nil, 50, Utils:Round(resWidth / 2), 0, widgetSettings, Rescan)
  self:AddSlider('frameOverviewMinHeight', "Min height", nil, 0, Utils:Round(resHeight / 2), 0, widgetSettings, Rescan)
  self:AddSlider('frameOverviewMaxHeight', "Max height", nil, 50, Utils:Round(resHeight / 2), 0, widgetSettings, Rescan)

  -- appearance
  widgetSettings.posX = widgetSettings.posX + 200
  widgetSettings.posY = DEFAULT_WIDGET_SETTINGS.posY
  self:AddHeadline("Appearance", widgetSettings)
  self:AddCheckButton('frameOverviewShowBorder', "Show border", nil, widgetSettings, UpdateFrameOverview)
  self:AddCheckButton('frameOverviewDisplayUnitOrigin', "Show how units were found", nil, widgetSettings, Rescan)

  -- test utils
  AddTestUtils(widgetSettings, true, false)

  ---------------------------
  --  Appearance Mouseover
  ---------------------------

  _, widgetSettings = self:AddSubCategory("  Mouseover Tooltip")

  -- pos
  self:AddHeadline("Positioning", widgetSettings)
  self:AddSlider('frameMouseoverOffsetX', "Offset X", nil, sliderMinCenteredFrameX, sliderMaxCenteredFrameX, 0, widgetSettings, UpdateFrameMouseover)
  self:AddSlider('frameMouseoverOffsetY', "Offset Y", nil, sliderMinCenteredFrameY, sliderMaxCenteredFrameY, 0, widgetSettings, UpdateFrameMouseover)
  self:AddDropdown('frameMouseoverStrata', "Layer", nil, dropdownMenuFrameStrata, 100, false, widgetSettings, UpdateFrameMouseover)

  -- size
  widgetSettings.posX = widgetSettings.posX + 200
  widgetSettings.posY = DEFAULT_WIDGET_SETTINGS.posY
  self:AddHeadline("Size", widgetSettings)
  self:AddSlider('frameMouseoverScale', "Scale", nil, 0.5, 2, 2, widgetSettings, UpdateFrameMouseover)

  -- appearance
  widgetSettings.posX = widgetSettings.posX + 200
  widgetSettings.posY = DEFAULT_WIDGET_SETTINGS.posY
  self:AddHeadline("Appearance", widgetSettings)
  self:AddCheckButton('frameMouseoverShowBorder', "Show border", nil, widgetSettings, UpdateFrameMouseover)

  -- test utils
  AddTestUtils(widgetSettings, true, false)

  ---------------------------
  --        Sounds
  ---------------------------

  _, widgetSettings = self:AddSubCategory("Sounds")
  self:AddHeadline("Sounds", widgetSettings)

  local function AddSoundSettings(aPartialSetting, aDisplayName) 
    posY = widgetSettings.posY
    self:AddCheckButton('sound'..aPartialSetting..'Enabled', aDisplayName, nil, widgetSettings)

    widgetSettings.posX = widgetSettings.posX + 150
    widgetSettings.posY = posY + 5
    self:AddDropdown('sound'..aPartialSetting..'Channel', "Channel", nil, dropdownMenuSoundChannel, 100, false, widgetSettings)

    widgetSettings.posX = widgetSettings.posX + 150
    widgetSettings.posY = posY + 5
    self:AddSlider('sound'..aPartialSetting..'MinTimeGap', "Minum time gap", nil, 0.1, 10, 2, widgetSettings)

    widgetSettings.posX = widgetSettings.posX + 180
    widgetSettings.posY = posY - 5
    self:AddButton("Play", 50, 150, 30, widgetSettings, function() AT.Main:SoundSpellFoundPlay(aPartialSetting, true) end)

    widgetSettings.posX = DEFAULT_WIDGET_SETTINGS.posX
    widgetSettings.posY = widgetSettings.posY + 5
    self:AddEditBox('sound'..aPartialSetting..'File', "Sound file", nil, false, 535, widgetSettings)

    -- add extra space
    widgetSettings.posY = widgetSettings.posY - 25
  end

  AddSoundSettings('Target', "Target")
  AddSoundSettings('Focus', "Focus")
  AddSoundSettings('Mouseover', "Mouseover tooltip")
  AddSoundSettings('Np', "Nameplate")
  AddSoundSettings('Overview', "Overview frame")

  ---------------------------
  --       Profiles
  ---------------------------

  self.frameProfilesHelp, widgetSettings = self:AddSubCategory("Profiles")
  self:AddHeadline("Profile Settings", widgetSettings)
  self:AddCheckButton('printLoadedProfileAtStartup', "Show loaded profiles at startup", "The setting is only applies if more than one profile exists for a profile type.", widgetSettings)
  widgetSettings.posY = widgetSettings.posY - 20
  self:AddHeadline("Profile Usage", widgetSettings)
  self:AddText(Strings.descriptionProfiles, 13, 582, widgetSettings)

  ---------------------------
  --     Experimental
  ---------------------------

  _, widgetSettings = self:AddSubCategory("Experimental")
  self:AddHeadline("Experimental / not finished features", widgetSettings)
  self:AddText(Strings.descriptionExperimentalSettings, 12, 582, widgetSettings)
  widgetSettings.posY = widgetSettings.posY - 20

  -- nameplates
  self:AddHeadline("Nameplates", widgetSettings)
  self:AddCheckButton('npIconKeepOpaque', "Keep icons always opaque", nil, widgetSettings, UpdateNameplates)
end

function FrameSettings:InitNonMage()
    -- create main frame
    self:EnsureMainFrame()
    local widgetSettings = Utils:TableShallowCopy(DEFAULT_WIDGET_SETTINGS)
    widgetSettings.parent = self.frame

    -- add content
    self:AddHeadline("Spellsteal Master", widgetSettings)
    self:AddText(Strings.descriptionDisabledForNonMages, 12, 582, widgetSettings)
    widgetSettings.posX = widgetSettings.posX - 2
    widgetSettings.posY = widgetSettings.posY - 20
    local cb = self:AddCheckButton(nil, "Always load for this class", nil, widgetSettings)
    cb:SetChecked(true)
    self:AddButton("Load Addon", 0, 400, 24, widgetSettings, function()
      -- add class to whitelist
      if cb:GetChecked() then
        AT.Main:AddCurrentClassToWhitelist()
      end
      -- init addon
      AT.Main:Setup(true)
      -- reopen options panel
      InterfaceOptionsFrame:Hide()
      HideUIPanel(GameMenuFrame)
      InterfaceOptionsFrame_OpenToCategory(self.frame)
    end)
end

function FrameSettings:OnUpdate()
  -- handle headline frames
  for _, v in ipairs(self.framePairs) do
    if v[1]:IsVisible() then
      -- switch category
      InterfaceOptionsFrame_OpenToCategory(v[2])
    end
  end

  -- unregister OnUpdate if settings closed
  if not InterfaceOptionsFrame:IsShown() then
    self:SetScript('OnUpdate', nil)
  end
end

------------------
--   Category
------------------

function FrameSettings:EnsureMainFrame()
  if self.frame then
    return
  end

  self.frame = CreateFrame('Frame', addonName..'sMain', UIParent)
  self.frame.name = self.categoryName 
  self.frame:Hide()

  InterfaceOptions_AddCategory(self.frame)
end

-- @RETURNS Frame
-- @RETURNS WidgetSettings
function FrameSettings:AddSubCategory(aDisplayName, --[[optional]] aLinkToPrevious)
  -- create frame
  local f = CreateFrame('Frame', nil, self.frame)
  f.name = aDisplayName
  f.parent = self.frame.name
  -- f:SetBackdrop({
  --   bgFile = 'Interface/Tooltips/UI-Tooltip-Background',
  --   -- edgeFile = 'Interface/Tooltips/UI-Tooltip-Border',
  --   tile = true, tileSize = 16, edgeSize = 16,
  --   insets = { left = 4, right = 4, top = 4, bottom = 4 }
  -- })
  -- f:SetBackdropColor(0, 0, 0, 1)
  InterfaceOptions_AddCategory(f)

  if not self.frame.childPanels then
    self.frame.childPanels = {}
  end
  tinsert(self.frame.childPanels, f)

  -- link to previous one
  if aLinkToPrevious then
    local prev = #self.frame.childPanels > 1 and self.frame.childPanels[#self.frame.childPanels - 1] or self.frame
    tinsert(self.framePairs, { prev, f })
  end

  f:Hide()

  --create widget settings from default settings, but with this frame as parent
  widgetSettings = Utils:TableShallowCopy(DEFAULT_WIDGET_SETTINGS)
  widgetSettings.parent = f

  return f, widgetSettings
end

------------------
--    Button
------------------

function FrameSettings:AddButton(aText, minWidth, maxWidth, Height, aWidgetSettings, aCallback, aIncreasePosY, aPosXOffset)
  local id = self:MakeFrameId()
  if aPosXOffset == nil then
    aPosXOffset = 0
  end

  -- create frame
  local frame = CreateFrame('Button', id, aWidgetSettings.parent, 'UIPanelButtonTemplate')
  frame:SetPoint(aWidgetSettings.point, (aWidgetSettings.posX + aPosXOffset), aWidgetSettings.posY)
  frame:SetScale(aWidgetSettings.scale)
  frame:SetText(aText)

  -- set size
  local strW = frame:GetTextWidth()
  local width
  if minWidth ~= nil and maxWidth ~= nil then
    width = Utils:Clamp(minWidth, strW + 15, maxWidth)
  else
    if DEBUG then
      error('not implemented')
    end
  end
  frame:SetSize(width, Height)

  -- setup callback
  frame:SetScript('OnClick', function()
      _assert(aCallback)
      aCallback()
  end)

  -- increase pos for next frame
  self:MayIncreasePosY(aIncreasePosY, aWidgetSettings, frame)

  self:OnFrameCreated(frame)
  return frame
end

------------------
--  CheckButton
------------------

function FrameSettings:AddCheckButton(aSetting, aText, aTooltip, aWidgetSettings, aCallback, aIncreasePosY, aPosXOffset)
  local id = self:MakeFrameId(aSetting)
  if aPosXOffset == nil then
    aPosXOffset = 0
  end

  -- create frame
  local frame = CreateFrame('CheckButton', id, aWidgetSettings.parent, 'ChatConfigCheckButtonTemplate')
  frame:SetPoint(aWidgetSettings.point, aWidgetSettings.posX + aPosXOffset, aWidgetSettings.posY)
  frame:SetScale(aWidgetSettings.scale)
  frame:SetScript('OnClick', 
    function()
      if aSetting then
        Settings[aSetting] = frame:GetChecked()
      end
      InvokeCallback(aCallback, frame)
    end
  )

  -- setup refresh function
  frame.RefreshValue = function(self)
    if aSetting then
      self:SetChecked(Settings[aSetting])
    end
  end
  frame:RefreshValue()

  -- set text
  if not Utils:StrIsNilOrEmpty(aText) then
    getglobal(frame:GetName().. 'Text'):SetText(aText)
  end

  -- set tooltip
  if not Utils:StrIsNilOrEmpty(aTooltip) then
    frame.tooltip = aTooltip
  end

  -- increase pos for next frame
  self:MayIncreasePosY(aIncreasePosY, aWidgetSettings, frame)

  self:OnFrameCreated(frame)
  return frame
end

------------------
--    Slider
------------------

function FrameSettings:AddSlider(aSetting, aText, aTooltip, aMinVal, aMaxVal, aDecimals, aWidgetSettings, aCallback, aIncreasePosY, aPosXOffset)
  local id = self:MakeFrameId(aSetting)
  if aPosXOffset == nil then
    aPosXOffset = 0
  end

  -- add extra space
  aWidgetSettings.posY = aWidgetSettings.posY - 15

  -- create frame for slider
  local slider = CreateFrame('Slider', id..'sl', aWidgetSettings.parent, 'OptionsSliderTemplate')
  slider:SetPoint(aWidgetSettings.point, aWidgetSettings.posX + aPosXOffset + 5, aWidgetSettings.posY)
  slider:SetScale(aWidgetSettings.scale)
  slider:SetMinMaxValues(aMinVal, aMaxVal)

  -- create frame for edit box
  slider.editBox = CreateFrame('EditBox', id..'eb', slider, 'InputBoxTemplate')
  slider.editBox:SetSize(50, 30)
  slider.editBox:ClearAllPoints()
  slider.editBox:SetPoint('TOP', slider, 'BOTTOM', 0, 5)
  slider.editBox:SetAutoFocus(false)
  slider.editBox:SetNumeric()

  -- setup refresh function
  slider.RefreshValue = function(self)
    if aSetting then
      local val = Settings[aSetting]
      self:SetValue(val)
      self.editBox:SetNumber(val)
    end
    self.editBox:SetCursorPosition(0)
  end
  slider:RefreshValue()

  -- setup scripts
  slider:SetScript('OnValueChanged', function(self, aValue)
    local val = Utils:Round(aValue, aDecimals)
    self.editBox:SetNumber(val)
    if aSetting then
      Settings[aSetting] = val
    end
    InvokeCallback(aCallback, slider)
  end)
  slider.editBox:SetScript('OnEnterPressed', function(self)
    local val = Utils:Round(self:GetNumber(), aDecimals)
    self:GetParent():SetValue(val)
    self:ClearFocus()
    self:SetNumber(val)
  end)

  -- override slider texts
  getglobal(slider:GetName()..'Low'):SetText(aMinVal)
  getglobal(slider:GetName()..'High'):SetText(aMaxVal)
  getglobal(slider:GetName()..'Text'):SetText(aText)

  -- set tooltip
  if not Utils:StrIsNilOrEmpty(aTooltip) then
    slider.tooltipText = aTooltip
  end

  -- increase pos for next frame
  self:MayIncreasePosY(aIncreasePosY, aWidgetSettings, slider)
  self:MayIncreasePosY(aIncreasePosY, aWidgetSettings, slider.editBox)

  self:OnFrameCreated(slider)
  return slider
end

------------------
--   EditBox
------------------

function FrameSettings:AddEditBox(aSetting, aText, aTooltip, aIsNumeric, aWidth, aWidgetSettings, aCallback, aIncreasePosY, aPosXOffset)
  local id = self:MakeFrameId(aSetting)
  if aPosXOffset == nil then
    aPosXOffset = 0
  end

  -- add extra space
  aWidgetSettings.posY = aWidgetSettings.posY - 15

  -- create frame
  local editBox = CreateFrame('EditBox', id, aWidgetSettings.parent, 'InputBoxTemplate')
  editBox:ClearAllPoints()
  editBox:SetPoint(aWidgetSettings.point, aWidgetSettings.posX + aPosXOffset + 7, aWidgetSettings.posY)
  editBox:SetSize(aWidth, 30)
  editBox:SetScale(aWidgetSettings.scale)
  editBox:SetAutoFocus(false)
  if aIsNumeric then
    editBox:SetNumeric()
  end

  -- setup scripts
  local function SaveValue(self)
    if aSetting then
      Settings[aSetting] = aIsNumeric and Utils:Round(self:GetNumber()) or self:GetText()
    end
    self:ClearFocus()
    InvokeCallback(aCallback, editBox)
  end
  editBox:SetScript('OnEnterPressed', SaveValue)
  editBox:SetScript('OnEditFocusLost', SaveValue)

  -- setup refresh function
  editBox.RefreshValue = function(self)
    if aSetting then
      if aIsNumeric then
        self:SetNumber(Utils:Round(Settings[aSetting]))
      else
        self:SetText(Settings[aSetting])
      end
    end
    self:SetCursorPosition(0)
  end
  editBox:RefreshValue()

  -- add fontstring
  editBox.fontString = editBox:CreateFontString(id..'fs', 'OVERLAY', GameFontNormal)
  editBox.fontString:SetPoint('TOPLEFT', editBox, -3, 10)
  editBox.fontString:SetFont('Fonts\\FRIZQT__.TTF', 12)
  editBox.fontString:SetText(aText)

  -- set tooltip
  -- TODO: doesn't work
  -- if not Utils:StrIsNilOrEmpty(aTooltip) then
  --   editBox.fontString.tooltip = aTooltip
  -- end

  -- increase pos for next frame
  self:MayIncreasePosY(aIncreasePosY, aWidgetSettings, editBox)

  self:OnFrameCreated(editBox)
  return editBox
end

------------------
--   Multiline EditBox
------------------

function FrameSettings:AddMultilineEditBox(aSetting, aWidth, aHeight, aWidgetSettings, aCallback, aIncreasePosY, aPosXOffset)
  local id = self:MakeFrameId(aSetting)
  if aPosXOffset == nil then
    aPosXOffset = 0
  end

  -- create frame
  local frame = CreateFrame('ScrollFrame', id, aWidgetSettings.parent, 'InputScrollFrameTemplate')
  -- workaround for an issue where this would be nil in Interface\SharedXML\SharedUIPanelTemplates.lua:674
  -- 2019-03-02, Patch 8.1
  frame.EditBox.cursorOffset = 0
  frame:ClearAllPoints()
  frame:SetPoint(aWidgetSettings.point, aWidgetSettings.posX + aPosXOffset, aWidgetSettings.posY)
  frame:SetSize(aWidth, aHeight)
  frame:SetScale(aWidgetSettings.scale)

  frame.CharCount:Hide()

  frame.EditBox:SetWidth(frame:GetWidth())
  frame.EditBox:SetFontObject('ChatFontNormal')
  frame.EditBox:SetScript('OnEditFocusLost', function(self)
    if aSetting then
      Settings[aSetting] = self:GetText()
    end
    self:ClearFocus()
    InvokeCallback(aCallback, frame)
  end)

  --setup refresh function
  frame.RefreshValue = function(self)
    if aSetting then
      self.EditBox:SetText(Settings[aSetting])
    end
  end
  frame:RefreshValue()

  -- increase pos for next frame
  self:MayIncreasePosY(aIncreasePosY, aWidgetSettings, frame)

  self:OnFrameCreated(frame)
  return frame
end

------------------
--   Dropdown
------------------

function FrameSettings:AddDropdown(aSetting, aText, aTooltip, aMenuList, aWidth, aIsBoolean, aWidgetSettings, aCallback, aIncreasePosY, aPosXOffset)
  -- search in the MenuList for the display name of a value
  local function getSettingValueDisplayName(aMenuList, aValue, aIsBoolean)
    if aIsBoolean and type(aValue) == 'boolean' then
      aValue = aValue and 1 or 0
    end
    -- "enum" value
    for _, entry in pairs(aMenuList) do
      if entry.value == aValue then
        return entry.text
      end
    end
    return ''
  end

  local id = self:MakeFrameId(aSetting)
  if aPosXOffset == nil then
    aPosXOffset = 0
  end

  -- add extra space
  aWidgetSettings.posY = aWidgetSettings.posY - 15

  -- create dropdown frame
  local dropdown = CreateFrame('Frame', id, aWidgetSettings.parent, 'UIDropDownMenuTemplate')
  dropdown:SetPoint(aWidgetSettings.point, aWidgetSettings.posX + aPosXOffset - 15, aWidgetSettings.posY)
  dropdown:SetScale(aWidgetSettings.scale)
  UIDropDownMenu_SetWidth(dropdown, aWidth)
  UIDropDownMenu_SetText(dropdown, getSettingValueDisplayName(aMenuList, Settings[aSetting], aIsBoolean))

  -- add fontstring
  dropdown.fontString = dropdown:CreateFontString(id..'fs', 'OVERLAY', GameFontNormal)
  dropdown.fontString:SetPoint('TOPLEFT', dropdown, 20, 10)
  dropdown.fontString:SetFont('Fonts\\FRIZQT__.TTF', 12)
  dropdown.fontString:SetText(aText)

  -- setup refresh function
  dropdown.RefreshValue = function(self)
    if aSetting then
      UIDropDownMenu_SetText(self, getSettingValueDisplayName(aMenuList, Settings[aSetting], aIsBoolean))
    end
  end

  -- initialize dropdown
  UIDropDownMenu_Initialize(dropdown, function()
    -- iterate over all entries in the menuList
    for _, entry in pairs(aMenuList) do
      -- create dropdown ready table
      local info = UIDropDownMenu_CreateInfo()
      -- and copy all infos into it
      info = Utils:TableShallowCopy(entry)
      -- setup OnClick
      info.func = function(self, aArg1, aArg2, aChecked)
        -- save setting
        if aSetting then
          Settings[aSetting] = Utils:Ternary(aIsBoolean, Utils:Ternary(self.value == 1, true, false), self.value)
        end
        -- refresh text
        UIDropDownMenu_SetText(dropdown, getSettingValueDisplayName(aMenuList, self.value))

        -- callback
        InvokeCallback(aCallback, dropdown)
      end

      UIDropDownMenu_AddButton(info)
    end
  end, '', 1, aWidth)

  -- set tooltip
  if not Utils:StrIsNilOrEmpty(aTooltip) then
    dropdown:SetScript('OnEnter', function(self)
      GameTooltip:SetOwner(self, 'ANCHOR_TOP')
      GameTooltip:SetText(aTooltip, nil, nil, nil, nil, true)
      GameTooltip:Show()
    end)
    dropdown:SetScript('OnLeave', function(self)
      GameTooltip:Hide()
    end)
  end

  -- increase pos for next frame
  self:MayIncreasePosY(aIncreasePosY, aWidgetSettings, dropdown)

  self:OnFrameCreated(dropdown)
  return dropdown
end

------------------
--   ColorPicker
------------------

function FrameSettings:AddColorPicker(aSetting, aText, aTooltip, aShowAlpha, aWidgetSettings, aCallback, aIncreasePosY, aPosXOffset)
  local id = self:MakeFrameId(aSetting)
  if aPosXOffset == nil then
    aPosXOffset = 0
  end

  -- add extra space
  aWidgetSettings.posY = aWidgetSettings.posY - 15

  -- create container for editboxes and button
  local frame = CreateFrame('Frame', id, aWidgetSettings.parent)
  frame:ClearAllPoints()
  frame:SetPoint(aWidgetSettings.point, aWidgetSettings.posX + aPosXOffset, aWidgetSettings.posY)
  frame:SetScale(aWidgetSettings.scale)

  -- add fontstring
  frame.fontString = frame:CreateFontString(id..'fs', 'OVERLAY', GameFontNormal)
  frame.fontString:SetPoint('TOPLEFT', frame, 5, 10)
  frame.fontString:SetFont('Fonts\\FRIZQT__.TTF', 12)
  frame.fontString:SetText(aText)

  local x = 7

  -- create pick color button
  frame.btn = CreateFrame('Button', id..'btn', frame, 'UIPanelButtonTemplate')
  frame.btn:SetText("Pick")
  frame.btn:SetPoint('LEFT', x - 5, 0)
  frame.btn:SetSize(frame.btn:GetTextWidth() + 15, 22)
  -- open color picker
  frame.btn:SetScript('OnClick', function()   
    -- setup callbacks
    local function OnChanged(aRestore)
      local r, g, b, a
      if not aRestore then
        -- color has changed
        r, g, b = ColorPickerFrame:GetColorRGB()
        a = OpacitySliderFrame:GetValue()
      else
        -- user hit cancel button
        r, g, b, a = unpack(aRestore)
      end
      -- round and clamp values
      r = Utils:Clamp(0, Utils:Round(r, 2), 1)
      g = Utils:Clamp(0, Utils:Round(g, 2), 1)
      b = Utils:Clamp(0, Utils:Round(b, 2), 1)
      a = Utils:Clamp(0, Utils:Round(1 - a, 2), 1)
      -- save
      if aSetting then
        Settings[aSetting] = { r = r, g = g, b = b, a = a }
      end
      -- callback
      InvokeCallback(aCallback, frame)
      -- refresh editboxes
      for _, eb in pairs(frame.editBoxes) do
        eb:RefreshValue()
      end
    end
    ColorPickerFrame.func = OnChanged
    ColorPickerFrame.opacityFunc = OnChanged
    ColorPickerFrame.cancelFunc = OnChanged

    -- setup rest and show
    local color = Settings[aSetting]
    ColorPickerFrame.hasOpacity = aShowAlpha
    ColorPickerFrame.opacity = 1 - color.a
    ColorPickerFrame.previousValues = { color.r, color.g, color.b, color.a };
    ColorPickerFrame:SetColorRGB(color.r, color.g, color.b)
    ColorPickerFrame:Hide() -- required to run the OnShow handler.
    ColorPickerFrame:Show()
  end)

  x = x + frame.btn:GetWidth() + 5

  -- create edit boxes
  local function AddEditBox(aTargetVal)
    -- create frame
    local editBox = CreateFrame('EditBox', id..'eb'..aTargetVal, frame, 'InputBoxTemplate')
    editBox:ClearAllPoints()
    editBox:SetPoint('LEFT', x, 0)
    editBox:SetSize(30, 30)
    editBox:SetAutoFocus(false)
    editBox:SetNumeric()

    -- setup scripts
    local function SaveValue(self)
      local val = Utils:Clamp(0, Utils:Round(self:GetNumber(), 2), 1)
      Settings[aSetting][aTargetVal] = val
      self:SetNumber(val)
      self:ClearFocus()
      InvokeCallback(aCallback, frame)
    end
    editBox:SetScript('OnEnterPressed', SaveValue)
    editBox:SetScript('OnEditFocusLost', SaveValue)

    -- setup refresh function
    editBox.RefreshValue = function(self)
      self:SetNumber(Settings[aSetting][aTargetVal])
      self:SetCursorPosition(0)
    end

    -- add fontstring
    editBox.fontString = frame:CreateFontString(id..'eb'..aTargetVal..'fs', 'OVERLAY', GameFontNormal)
    editBox.fontString:SetPoint('BOTTOM', editBox, 0, -3)
    editBox.fontString:SetFont('Fonts\\FRIZQT__.TTF', 10)
    editBox.fontString:SetTextColor(0.8, 0.8, 0.8, 1)
    editBox.fontString:SetText(aTargetVal)

    -- attach to main frame
    if not frame.editBoxes then
      frame.editBoxes = {}
    end
    tinsert(frame.editBoxes, editBox)

    -- increase x for next editbox
    x = x + editBox:GetWidth() + 10
  end
  AddEditBox('r')
  AddEditBox('g')
  AddEditBox('b')
  if aShowAlpha then
    AddEditBox('a')
  end

  -- setup refresh function
  frame.RefreshValue = function(self)
    for _, f in ipairs(frame.editBoxes) do
      f:RefreshValue()
    end
  end
  frame:RefreshValue()

  -- adjust size of parent frame
  frame:SetSize(x + frame.editBoxes[1]:GetWidth(), 30)
  frame:Show()

  -- increase pos for next frame
  self:MayIncreasePosY(aIncreasePosY, aWidgetSettings, frame, 5)

  self:OnFrameCreated(frame)
  return frame
end

------------------
-- FontString / Headline
------------------

function FrameSettings:AddHeadline(aText, aWidgetSettings, aIncreasePosY)
  -- create frame
  local fs = aWidgetSettings.parent:CreateFontString(nil, 'OVERLAY', GameFontNormal)
  fs:SetPoint(aWidgetSettings.point, aWidgetSettings.posX, aWidgetSettings.posY)
  fs:SetScale(aWidgetSettings.scale)
  fs:SetFont('Fonts\\FRIZQT__.TTF', 18, 'OUTLINE')
  fs:SetTextColor(unpack(Constants.COLOR_YELLOW_RGB))
  fs:SetText(aText)

  -- increase pos for next frame
  self:MayIncreasePosY(aIncreasePosY, aWidgetSettings, fs, 10)

  self:OnFrameCreated(fs)
  return fs
end

------------------
-- FontString / Headline
------------------

function FrameSettings:AddText(aText, aFontSize, aWidth, aWidgetSettings, aIncreasePosY)
  -- create frame
  local fs = aWidgetSettings.parent:CreateFontString(nil, 'OVERLAY', GameFontNormal)
  fs:SetScale(aWidgetSettings.scale)
  fs:SetWidth(aWidth)
  fs:SetPoint(aWidgetSettings.point, aWidgetSettings.posX, aWidgetSettings.posY)
  fs:SetJustifyV('TOP')
  fs:SetJustifyH('LEFT')
  fs:SetFont('Fonts\\FRIZQT__.TTF', aFontSize, 'OUTLINE')
  fs:SetText(aText)

  -- increase pos for next frame
  self:MayIncreasePosY(aIncreasePosY, aWidgetSettings, fs, 10)

  self:OnFrameCreated(fs)
  return fs
end

------------------
--     Misc
------------------

-- opens profiles commands help
-- used weird func name, so that it can get called data driven via slash commands
function FrameSettings:ProfilesHelp()
  InterfaceOptionsFrame_OpenToCategory(self.frameProfilesHelp)
end

------------------
--    Shared
------------------

function FrameSettings:MakeFrameId(aName)
  return addonName..'fs'..(aName and aName..#self.createdFrames or #self.createdFrames)
end

function FrameSettings:OnFrameCreated(aFrame)
  self.createdFrames[#self.createdFrames + 1] = aFrame
end

-- @RETURNS Frame
function FrameSettings:LastCreatedFrame()
  return self.createdFrames[#self.createdFrames]
end

function FrameSettings:RefreshAllValues()
  for _, v in ipairs(self.createdFrames) do
    if v.RefreshValue ~= nil then
      v:RefreshValue()
    end
  end
end

function FrameSettings:MayIncreasePosY(aIncreasePosY, aWidgetSettings, aFrame, aExtraSpace)
  if aIncreasePosY == true or aIncreasePosY == nil then
    aWidgetSettings.posY = aWidgetSettings.posY - aFrame:GetHeight() - (aExtraSpace or 0)
  end
end
