local addonName, AT = ...
local DEBUG = false

-- file imports
local Constants = AT.Constants
local Utils = AT.Utils

-- enum imports
local PROFILE_TYPE = AT.Enums.PROFILE_TYPE

--[[
  NOTES:
    - "global" settings are not affected by profiles
    - profile management funcs are called dynamically from Commands.lua
    - setting names must be unique across all setting types
]]

local DEFAULT_CHARACTER_SETTINGS = {
  currentProfileFilter = Constants.DEFAULT_PROFILE,
  currentProfileGeneral = Constants.DEFAULT_PROFILE
}

local DEFAULT_GLOBAL_SETTINGS = {
  whitelistedClasses = Constants.WHITELISTED_CLASSES,
  showWelcomeMsg = true,
  printLoadedProfileAtStartup = true,

  testFramesEnabled = false,
  testFramesShowMouseover = true,
  testFramesShowNameplateIcons = false
}

local DEFAULT_GENERAL_SETTINGS = {
  -- general
  scanOutOfCombat = true,
  buffStolenChatMessage = 1,
  frameDragWithModifierKey = false,

  -- sounds
  soundTargetEnabled = true,
  soundTargetFile = 'Interface\\AddOns\\SpellstealMaster\\sounds\\371753__problematist__dhumm2.ogg',
  soundTargetChannel = 'SFX',
  soundTargetMinTimeGap = 1,
  soundFocusEnabled = true,
  soundFocusFile = 'Interface\\AddOns\\SpellstealMaster\\sounds\\371753__problematist__dhumm2.ogg',
  soundFocusChannel = 'SFX',
  soundFocusMinTimeGap = 1,
  soundMouseoverEnabled = true,
  soundMouseoverFile = 'Interface\\AddOns\\SpellstealMaster\\sounds\\262958__unicornaphobist__glass-tap.wav',
  soundMouseoverChannel = 'SFX',
  soundMouseoverMinTimeGap = 0.1,
  soundOverviewEnabled = true,
  soundOverviewFile = 'Interface\\AddOns\\SpellstealMaster\\sounds\\371752__problematist__dhumm.ogg',
  soundOverviewChannel = 'SFX',
  soundOverviewMinTimeGap = 2.5,
  soundNpEnabled = true,
  soundNpFile = 'Interface\\AddOns\\SpellstealMaster\\sounds\\371752__problematist__dhumm.ogg',
  soundNpChannel = 'SFX',
  soundNpMinTimeGap = 2.5,

  -- appearance
  colorBg = { r = 0, g = 0, b = 0, a = 0.55 },
  colorBorder = { r = 0.18, g = 0.65, b = 1, a = 1 },
  colorHeadline = { r = 0.18, g = 0.65, b = 1, a = 1 },
  colorUnitName = { r = 0.98, g = 0.72, b = 0.72, a = 1 },
  colorUnitOrigin = { r = 0.52, g = 0.52, b = 0.52, a = 1 },
  colorBuffName = { r = 0.58, g = 1, b = 0.68, a = 1 },
  colorBuffCount = { r = 0.69, g = 0.69, b = 0.69, a = 1 },
  colorBuffExpTime = { r = 0.69, g = 0.69, b = 0.69, a = 1 },
  alphaNoMana = 0.33,
  alphaOutOfRange = 0.66,

  -- appearance frame target
  frameTargetEnabled = true,
  frameTargetShowBorder = true,
  frameTargetStrata = 'MEDIUM',
  frameTargetScale = 1,
  frameTargetPivot = 'CENTER',
  frameTargetAnchor = 'CENTER',
  frameTargetOffsetX = 0,
  frameTargetOffsetY = 100,

  -- appearance frame focus
  frameFocusEnabled = true,
  frameFocusShowBorder = true,
  frameFocusStrata = 'MEDIUM',
  frameFocusScale = 1,
  frameFocusPivot = 'CENTER',
  frameFocusAnchor = 'CENTER',
  frameFocusOffsetX = 200,
  frameFocusOffsetY = 0,

  -- appearance frame overview
  frameOverviewEnabled = true,
  frameOverviewShowBorder = true,
  frameOverviewStrata = 'MEDIUM',
  frameOverviewScale = 1,
  frameOverviewMinWidth = 0,
  frameOverviewMaxWidth = 150,
  frameOverviewMinHeight = 0,
  frameOverviewMaxHeight = 200,
  frameOverviewPivot = 'CENTER',
  frameOverviewAnchor = 'CENTER',
  frameOverviewOffsetX = -200,
  frameOverviewOffsetY = 0,
  frameOverviewDisplayUnitOrigin = true,

  -- appearance frame mousover
  frameMouseoverEnabled = true,
  frameMouseoverShowBorder = true,
  frameMouseoverStrata = 'TOOLTIP',
  frameMouseoverScale = 1,
  frameMouseoverOffsetX = 20,
  frameMouseoverOffsetY = 0,

  -- appearance nameplates
  npEnabled = true,
  npStrata = 'MEDIUM',
  npOffsetX = 0,
  npOffsetY = 39,
  npIconWidth = 20,
  npIconHeight = 20,
  npIconBorderSize = 2,
  npIconBorderColor = { r = 1, g = 1, b = 0, a = 1 },
  npIconSpacing = 4,
  npIconOrientation = 1,
  npIconKeepOpaque = false
}

local DEFAULT_FILTER_SETTINGS = {
  -- filter unit
  filterIgnoreUnitsNotAffectingCombat = false,

  -- csv filtering
  filterIgnoreGroupBuffs = true,
  filterGlobalCsv = '',
  filterGlobalCsvIsBlacklist = true,
  filterContextualCsv = '',
  filterContextualCsvIsBlacklist = true,
  filterContextualLua = '',

  -- filter important
  filterImportantMinExpirationTime = 2,

  -- filter overview
  filterOverviewMinExpirationTime = 5,
  filterOverviewShowTarget = false,
  filterOverviewShowFocus = false,
  filterOverviewShowNameplates = true,
  filterOverviewShowBosses = true,
  filterOverviewShowPetTarget = true,
  filterOverviewShowGroupMemberTargets = true,
  filterOverviewShowGroupMemberPetTargets = true,
}

local Settings = {}
Settings.Character = {}
Settings.Global = {}
Settings.General = {}
Settings.Filter = {}
Settings.OnProfileChanged = {
  listeners = {},
  Add = function(self, aTable, aFunc, aProfileTypeFlags) 
    tinsert(self.listeners, {
      table = aTable,
      func = aFunc,
      flags = aProfileTypeFlags
    })
  end,
  Broadcast = function(self, aProfileType)
    for _, v in ipairs(self.listeners) do
      if Enums:HasFlag(v.flags, aProfileType) then
        v.func(v.table, aProfileType)
      end
    end
  end
}
Settings.isInitialized = false

AT.Settings = Settings

function Settings:Init()
  if self.isInitialized then
    return
  end

  -- set own metatable
  setmetatable(Settings, { __index = Settings.Get, __newindex = Settings.Set })

  -- character
  if SPELLSTEALMASTER_SETTINGS_CHARACTER == nil then
    SPELLSTEALMASTER_SETTINGS_CHARACTER = {}
  end
  setmetatable(SPELLSTEALMASTER_SETTINGS_CHARACTER, { __index = DEFAULT_CHARACTER_SETTINGS })
  self.Character = SPELLSTEALMASTER_SETTINGS_CHARACTER

  -- global
  self:GetGlobalSettings()

  -- general and filter
  self:LoadCurrentProfile(PROFILE_TYPE.GENERAL)
  self:LoadCurrentProfile(PROFILE_TYPE.FILTER)

  self.isInitialized = true
end

-- Allowes tp access global settings before addon is fully initialized. Doesn't load any profiles.
-- Used by the "load addon only for whitelisted classes" feature
-- @RETURNS Table
function Settings:GetGlobalSettings()
  if getmetatable(self.Global) == nil then
    if SPELLSTEALMASTER_SETTINGS_GLOBAL == nil then
      SPELLSTEALMASTER_SETTINGS_GLOBAL = {}
    end
    setmetatable(SPELLSTEALMASTER_SETTINGS_GLOBAL, { __index = DEFAULT_GLOBAL_SETTINGS })
    self.Global = SPELLSTEALMASTER_SETTINGS_GLOBAL
  end
  return self.Global
end

function Settings:Get(aKey)
  local val = self.General[aKey]
  if val == nil then
    val = self.Filter[aKey]
  end
  if val == nil then
    val = self.Global[aKey]
  end
  if val == nil then
    val = self.Character[aKey]
  end
  if DEBUG and val == nil then
    error('Setting '..aKey..' does not exist')
  end
  return val
end

function Settings:Set(aKey, aValue)
  if self.General[aKey] ~= nil then
    self.General[aKey] = aValue
  elseif self.Filter[aKey] ~= nil then
    self.Filter[aKey] = aValue
  elseif self.Global[aKey] ~= nil then
    self.Global[aKey] = aValue
  elseif self.Character[aKey] ~= nil then
    self.Character[aKey] = aValue
  elseif DEBUG then
    error('Setting '..aKey..' does not exist')
  end
end

function Settings:CreateProfile(aProfileType, aProfileName)
  -- global table
  local gt = self:GetGlobalTable(aProfileType)

  -- check if profile already exists
  if self:ProfileExists(aProfileType, aProfileName) then
    -- print error to chat
    Utils:Print(string.format("profile %s (type: %s) already exists", self:DyeProfileName(aProfileName), self:ProfileTypeToStrColored(aProfileType)))
    return
  end

  -- create profile
  gt.profiles[aProfileName] = {}
  -- print to chat
  Utils:Print(string.format("created profile %s (type: %s)", self:DyeProfileName(aProfileName), self:ProfileTypeToStrColored(aProfileType)))
  -- load profile
  self:LoadProfile(aProfileType, aProfileName)
end

-- @RETURNS Boolean
function Settings:CopyProfile(aProfileType, aFromProfileName, aToProfileName, --[[optional]] aLoadCreated)
  -- global table
  local gt = self:GetGlobalTable(aProfileType)

  -- check if from-profile exists
  if not self:ProfileExists(aProfileType, aFromProfileName, true) then
    return false
  end

  -- check if to-profile already exists
  if self:ProfileExists(aProfileType, aToProfileName) then
    -- print error to chat
    Utils:Print(string.format("profile %s (type: %s) already exists", self:DyeProfileName(aToProfileName), self:ProfileTypeToStrColored(aProfileType)))
    return false
  end

  -- copy profile
  gt.profiles[aToProfileName] = Utils:TableDeepCopy(gt.profiles[aFromProfileName])
  -- print to chat
  Utils:Print(string.format("created profile %s (type: %s) from profile %s",
    self:DyeProfileName(aToProfileName), self:ProfileTypeToStrColored(aProfileType), self:DyeProfileName(aFromProfileName)))
  -- load new created profile
  if aLoadCreated == nil or aLoadCreated == true then
    self:LoadProfile(aProfileType, aToProfileName)
  end

  return true
end

function Settings:RenameProfile(aProfileType, aFromProfileName, aToProfileName)
  if self:CopyProfile(aProfileType, aFromProfileName, aToProfileName, false) then
    self:DeleteProfile(aProfileType, aFromProfileName)
  end
end

function Settings:LoadProfile(aProfileType, aProfileName)
  -- get global table
  local gt = self:GetGlobalTable(aProfileType)

  -- check if profile exists
  if not self:ProfileExists(aProfileType, aProfileName, true) then
    return
  end

  -- save current
  self:SetCurrentProfileName(aProfileType, aProfileName)

  -- get profile table
  local pt = gt.profiles[aProfileName]
  setmetatable(pt, {__index = self:GetDefaultSettings(aProfileType)})

  -- set for public access
  if aProfileType == PROFILE_TYPE.GENERAL then
    self.General = pt
  else
    self.Filter = pt
  end

  -- print to chat if at least two profiles exists
  if self.isInitialized or (Utils:TableLen(gt.profiles) > 1 and Settings.printLoadedProfileAtStartup) then
    Utils:Print(string.format("loaded profile %s (type: %s)", self:DyeProfileName(aProfileName), self:ProfileTypeToStrColored(aProfileType)))
  end

  -- broadcast profile changed
  self.OnProfileChanged:Broadcast(aProfileType)
end

function Settings:LoadCurrentProfile(aProfileType)
  -- get global table
  local gt = self:GetGlobalTable(aProfileType)

  -- get current profile name
  local profileName = self:GetCurrentProfileName(aProfileType)

  -- if current profile does not exist
  if not self:ProfileExists(aProfileType, profileName) then
    -- first search for other profiles
    profiles = self:GetProfileNamesSorted(aProfileType)
    if #profiles > 0 then
      -- switch profile alphabetically sfirst profile
      profileName = profiles[1]
      self:SetCurrentProfileName(aProfileType, profileName)
    else
      -- otherwise switch to default profile
      profileName = Constants.DEFAULT_PROFILE
      self:SetCurrentProfileName(aProfileType, profileName)

      -- ensure default profile
      if not self:ProfileExists(aProfileType, profileName) then
        self:CreateProfile(aProfileType, profileName)
        return
      end
    end
  end

  -- load profile
  self:LoadProfile(aProfileType, profileName)
end

function Settings:LoadNextProfile(aProfileType)
  -- get profiles alphabetically sorted
  profiles = self:GetProfileNamesSorted(aProfileType)

  -- get next profile
  local currentIndex = Utils:TableValueIndex(profiles, self:GetCurrentProfileName(aProfileType))
  local nextProfile = profiles[currentIndex < #profiles and currentIndex + 1 or 1]

  -- load next profile
  self:LoadProfile(aProfileType, nextProfile)
end

function Settings:DeleteProfile(aProfileType, aProfileName)
  -- check if profile exists
  if not self:ProfileExists(aProfileType, aProfileName, true) then
    return
  end

  -- get global table
  local gt = self:GetGlobalTable(aProfileType)
  gt.profiles[aProfileName] = nil

  -- print to chat
  Utils:Print(string.format("deleted profile %s (type: %s)", self:DyeProfileName(aProfileName), self:ProfileTypeToStrColored(aProfileType)))

  -- check if the deleted one was the current one
  if aProfileName == self:GetCurrentProfileName(aProfileType) then
    -- load default profile
    self:LoadCurrentProfile(aProfileType)
  end
end

function Settings:ProfileExists(aProfileType, aProfileName, --[[optional]] aPrintOnError)
  -- get global table
  local gt = self:GetGlobalTable(aProfileType)

  if gt.profiles[aProfileName] then
    return true
  end

  if aPrintOnError then
    -- print error to chat
    Utils:Print(string.format("profile %s (type: %s) does not exists", self:DyeProfileName(aProfileName), self:ProfileTypeToStrColored(aProfileType)))
  end

  return false
end

function Settings:ListProfiles(aProfileType)
  local strBuilder = {}

  -- get general profiles
  if aProfileType == nil or aProfileType == PROFILE_TYPE.GENERAL then
    tinsert(strBuilder, '\n')
    tinsert(strBuilder, "General profiles:")
    for k, _ in pairs(self:GetGlobalTable(PROFILE_TYPE.GENERAL).profiles) do
      tinsert(strBuilder, '\n- '..k)
    end
  end

  -- get filter profiles
  if aProfileType == nil or aProfileType == PROFILE_TYPE.FILTER then
    tinsert(strBuilder, '\n')
    tinsert(strBuilder, "Filter profiles:")
    for k, _ in pairs(self:GetGlobalTable(PROFILE_TYPE.FILTER).profiles) do
      tinsert(strBuilder, '\n- '..k)
    end
  end

  -- print to chat
  Utils:Print(table.concat(strBuilder))
end

function Settings:GetCurrentProfileName(aProfileType)
  if aProfileType == PROFILE_TYPE.GENERAL then
    return self.currentProfileGeneral
  else
    return self.currentProfileFilter
  end
end

function Settings:SetCurrentProfileName(aProfileType, aProfileName)
  if aProfileType == PROFILE_TYPE.GENERAL then
    self.currentProfileGeneral = aProfileName
  else
    self.currentProfileFilter = aProfileName
  end
end

function Settings:GetProfileNamesSorted(aProfileType)
  -- create alphabetically sorted table
  local gt = self:GetGlobalTable(aProfileType)
  local names = {}
  for k, _ in pairs(gt.profiles) do
    tinsert(names, k)
  end
  table.sort(names, function(a, b) return a < b end)
  return names
end

function Settings:GetGlobalTable(aProfileType)
  local t
  -- get global table
  if aProfileType == PROFILE_TYPE.GENERAL then
    if SPELLSTEALMASTER_SETTINGS_GENERAL == nil then
      SPELLSTEALMASTER_SETTINGS_GENERAL = {}
    end
    t = SPELLSTEALMASTER_SETTINGS_GENERAL
  else
    if SPELLSTEALMASTER_SETTINGS_FILTER == nil then
      SPELLSTEALMASTER_SETTINGS_FILTER = {}
    end
    t = SPELLSTEALMASTER_SETTINGS_FILTER
  end

  -- ensure properties
  if t.profiles == nil then
    t.profiles = {}
  end

  return t
end

function Settings:GetDefaultSettings(aProfileType)
  return aProfileType == PROFILE_TYPE.GENERAL
    and DEFAULT_GENERAL_SETTINGS
    or DEFAULT_FILTER_SETTINGS
end

function Settings:ProfileTypeToStr(aProfileType)
  return aProfileType == PROFILE_TYPE.GENERAL
    and "general"
    or "filter"
end

function Settings:ProfileTypeToStrColored(aProfileType)
  return string.format("|c%s%s|r", Constants.COLOR_PROFILE_TYPE_HEX, self:ProfileTypeToStr(aProfileType))
end

function Settings:DyeProfileName(aProfileName)
  return string.format("|c%s%s|r", Constants.COLOR_PROFILE_NAME_HEX, aProfileName)
end
